{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Vlasov-Poisson \n",
    "\n",
    "In this notebook we present a multithreaded version of the Vlasov-Poisson system in 1D1D phase space.\n",
    "Equations are solved numerically using semi-lagrangian method. \n",
    "\n",
    " ## Semi-Lagrangian method\n",
    "\n",
    " Let us consider an abstract scalar advection equation of the form\n",
    " $$\n",
    " \\frac{\\partial f}{\\partial t}+ a(x, t) \\cdot \\nabla f = 0. \n",
    " $$\n",
    " The characteristic curves associated to this equation are the solutions of the ordinary differential equations\n",
    " $$\n",
    " \\frac{dX}{dt} = a(X(t), t)\n",
    " $$\n",
    " We shall denote by $X(t, x, s)$ the unique solution of this equation associated to the initial condition $X(s) = x$.\n",
    "\n",
    " The classical semi-Lagrangian method is based on a backtracking of characteristics. Two steps are needed to update the distribution function $f^{n+1}$ at $t^{n+1}$ from its value $f^n$ at time $t^n$ :\n",
    " 1. For each grid point $x_i$ compute $X(t^n; x_i, t^{n+1})$ the value of the characteristic at $t^n$ which takes the value $x_i$ at $t^{n+1}$.\n",
    " 2. As the distribution solution of first equation verifies \n",
    " $$f^{n+1}(x_i) = f^n(X(t^n; x_i, t^{n+1})),$$\n",
    " we obtain the desired value of $f^{n+1}(x_i)$ by computing $f^n(X(t^n;x_i,t^{n+1})$ by interpolation as $X(t^n; x_i, t^{n+1})$ is in general not a grid point.\n",
    "\n",
    " *[Eric Sonnendrücker - Numerical methods for the Vlasov equations](http://www-m16.ma.tum.de/foswiki/pub/M16/Allgemeines/NumMethVlasov/Num-Meth-Vlasov-Notes.pdf)*\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import sys\n",
    "\n",
    "if sys.platform == \"darwin\":\n",
    "    %env CC='gcc-10'"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Bspline\n",
    "\n",
    "The `bspline` function return the value at x in [0,1[ of the B-spline with \n",
    "integer nodes of degree p with support starting at j.\n",
    "Implemented recursively using the de Boor's recursion formula\n",
    "\n",
    "[De Boor's Algorithm - Wikipedia](https://en.wikipedia.org/wiki/De_Boor%27s_algorithm)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "from scipy.fftpack import fft, ifft \n",
    "\n",
    "def bspline_python(p, j, x):\n",
    "        \"\"\"Return the value at x in [0,1[ of the B-spline with \n",
    "        integer nodes of degree p with support starting at j.\n",
    "        Implemented recursively using the de Boor's recursion formula\"\"\"\n",
    "        assert (x >= 0.0) & (x <= 1.0)\n",
    "        assert (type(p) == int) & (type(j) == int)\n",
    "        if p == 0:\n",
    "            if j == 0:\n",
    "                return 1.0\n",
    "            else:\n",
    "                return 0.0\n",
    "        else:\n",
    "            w = (x - j) / p\n",
    "            w1 = (x - j - 1) / p\n",
    "        return w * bspline_python(p - 1, j, x) + (1 - w1) * bspline_python(p - 1, j + 1, x)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Advection:\n",
    "    \n",
    "    \"\"\" Class to compute BSL advection of 1d function \"\"\"\n",
    "    \n",
    "    def __init__(self, p, xmin, xmax, ncells, bspline):\n",
    "        assert p & 1 == 1  # check that p is odd\n",
    "        self.p = p\n",
    "        self.ncells = ncells\n",
    "        # compute eigenvalues of degree p b-spline matrix\n",
    "        self.modes = 2 * np.pi * np.arange(ncells) / ncells\n",
    "        self.deltax = (xmax - xmin) / ncells\n",
    "        self.bspline = bspline\n",
    "        self.eig_bspl = bspline(p, -(p + 1) // 2, 0.0)\n",
    "        for j in range(1, (p + 1) // 2):\n",
    "            self.eig_bspl += bspline(p, j - (p + 1) // 2, 0.0) * 2 * np.cos(j * self.modes)\n",
    "            \n",
    "        self.eigalpha = np.zeros(ncells, dtype=complex)\n",
    "    \n",
    "    def __call__(self, f, alpha):\n",
    "        \"\"\"compute the interpolating spline of degree p of odd degree \n",
    "        of a function f on a periodic uniform mesh, at\n",
    "        all points xi-alpha\"\"\"\n",
    "        p = self.p\n",
    "        assert (np.size(f) == self.ncells)\n",
    "        # compute eigenvalues of cubic splines evaluated at displaced points\n",
    "        ishift = np.floor(-alpha / self.deltax)\n",
    "        beta = -ishift - alpha / self.deltax\n",
    "        self.eigalpha.fill(0.)\n",
    "        for j in range(-(p-1)//2, (p+1)//2 + 1):\n",
    "            self.eigalpha += self.bspline(p, j-(p+1)//2, beta) * np.exp((ishift+j)*1j*self.modes)\n",
    "            \n",
    "        # compute interpolating spline using fft and properties of circulant matrices\n",
    "        return np.real(ifft(fft(f) * self.eigalpha / self.eig_bspl))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def interpolation_test(bspline):\n",
    "    \"\"\" Test to check interpolation\"\"\"\n",
    "    n = 64\n",
    "    cs = Advection(3,0,1,n, bspline)\n",
    "    x = np.linspace(0,1,n, endpoint=False)\n",
    "    f = np.sin(x*4*np.pi)\n",
    "    alpha = 0.2\n",
    "    return np.allclose(np.sin((x-alpha)*4*np.pi), cs(f, alpha))\n",
    "    \n",
    "\n",
    "interpolation_test(bspline_python)\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "If we profile the code we can see that a lot of time is spent in `bspline` calls"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Accelerate with Fortran"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%load_ext fortranmagic"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%fortran\n",
    "recursive function bspline_fortran(p, j, x) result(res)\n",
    "    integer :: p, j\n",
    "    real(8) :: x, w, w1\n",
    "    real(8) :: res\n",
    "\n",
    "    if (p == 0) then\n",
    "        if (j == 0) then\n",
    "            res = 1.0\n",
    "            return\n",
    "        else\n",
    "            res = 0.0\n",
    "            return\n",
    "        end if\n",
    "    else\n",
    "        w = (x - j) / p\n",
    "        w1 = (x - j - 1) / p\n",
    "    end if\n",
    "    \n",
    "    res = w * bspline_fortran(p-1,j,x) &\n",
    "    +(1-w1)*bspline_fortran(p-1,j+1,x)\n",
    "\n",
    "end function bspline_fortran"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "interpolation_test(bspline_fortran)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Numba\n",
    "\n",
    "Create a optimized function of bspline python function with Numba. Call it bspline_numba."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from numba import jit,  int32, float64\n",
    "from scipy.fftpack import fft, ifft\n",
    "\n",
    "@jit(\"float64(int32,int32,float64)\",nopython=True)\n",
    "def bspline_numba(p, j, x):\n",
    "    \n",
    "        \"\"\"Return the value at x in [0,1[ of the B-spline with \n",
    "        integer nodes of degree p with support starting at j.\n",
    "        Implemented recursively using the de Boor's recursion formula\"\"\"\n",
    "        \n",
    "        assert ((x >= 0.0) & (x <= 1.0))\n",
    "        if p == 0:\n",
    "            if j == 0:\n",
    "                return 1.0\n",
    "            else:\n",
    "                return 0.0\n",
    "        else:\n",
    "            w = (x-j)/p\n",
    "            w1 = (x-j-1)/p\n",
    "        return w * bspline_numba(p-1,j,x)+(1-w1)*bspline_numba(p-1,j+1,x)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "interpolation_test(bspline_numba)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Pythran"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import pythran\n",
    "\n",
    "%load_ext pythran.magic"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%pythran\n",
    "\n",
    "import numpy as np\n",
    "\n",
    "#pythran export bspline_pythran(int,int,float64)\n",
    "def bspline_pythran(p, j, x):\n",
    "    if p == 0:\n",
    "        if j == 0:\n",
    "            return 1.0\n",
    "        else:\n",
    "            return 0.0\n",
    "    else:\n",
    "        w = (x-j)/p\n",
    "        w1 = (x-j-1)/p\n",
    "    return w * bspline_pythran(p-1,j,x)+(1-w1)*bspline_pythran(p-1,j+1,x)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "interpolation_test(bspline_pythran)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Cython"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%load_ext cython"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%cython -a\n",
    "def bspline_cython(p, j, x):\n",
    "        \"\"\"Return the value at x in [0,1[ of the B-spline with \n",
    "        integer nodes of degree p with support starting at j.\n",
    "        Implemented recursively using the de Boor's recursion formula\"\"\"\n",
    "        assert (x >= 0.0) & (x <= 1.0)\n",
    "        assert (type(p) == int) & (type(j) == int)\n",
    "        if p == 0:\n",
    "            if j == 0:\n",
    "                return 1.0\n",
    "            else:\n",
    "                return 0.0\n",
    "        else:\n",
    "            w = (x - j) / p\n",
    "            w1 = (x - j - 1) / p\n",
    "        return w * bspline_cython(p - 1, j, x) + (1 - w1) * bspline_cython(p - 1, j + 1, x)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%cython\n",
    "import cython\n",
    "\n",
    "@cython.cdivision(True)\n",
    "cpdef double bspline_cython(int p, int j, double x):\n",
    "        \"\"\"Return the value at x in [0,1[ of the B-spline with \n",
    "        integer nodes of degree p with support starting at j.\n",
    "        Implemented recursively using the de Boor's recursion formula\"\"\"\n",
    "        cdef double w, w1\n",
    "        if p == 0:\n",
    "            if j == 0:\n",
    "                return 1.0\n",
    "            else:\n",
    "                return 0.0\n",
    "        else:\n",
    "            w = (x - j) / p\n",
    "            w1 = (x - j - 1) / p\n",
    "        return w * bspline_cython(p-1,j,x)+(1-w1)*bspline_cython(p-1,j+1,x)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "interpolation_test(bspline_cython)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%config InlineBackend.figure_format = 'retina'\n",
    "import matplotlib.pyplot as plt\n",
    "import numpy as np\n",
    "\n",
    "plt.rcParams['figure.figsize'] = (11,7)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from tqdm.notebook import tqdm\n",
    "from collections import defaultdict\n",
    "\n",
    "Mrange = (2 ** np.arange(6, 12)).astype(int)\n",
    "opts = [bspline_fortran, bspline_pythran, bspline_cython, bspline_numba]\n",
    "\n",
    "times = defaultdict(list)\n",
    "for M in tqdm(Mrange):\n",
    "    x = np.linspace(0, 1, M, endpoint=False)\n",
    "    for opt in opts:\n",
    "        f = np.sin(x*4*np.pi)\n",
    "        cs = Advection(5, 0, 1, M, opt)\n",
    "    \n",
    "        alpha = 0.1\n",
    "        ts = %timeit -oq -n 100 cs(f, alpha)\n",
    "        times[opt.__name__].append(ts.best)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "for o, t in times.items():\n",
    "    plt.loglog(Mrange, t, label=o)\n",
    "    \n",
    "plt.legend(loc='lower right')\n",
    "plt.xlabel('Number of points')\n",
    "plt.ylabel('Execution Time (s)');"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%cython\n",
    "import cython\n",
    "import numpy as np\n",
    "cimport numpy as np\n",
    "from scipy.fftpack import fft, ifft\n",
    "\n",
    "@cython.cdivision(True)\n",
    "cdef double bspline_cython(int p, int j, double x):\n",
    "        \"\"\"Return the value at x in [0,1[ of the B-spline with \n",
    "        integer nodes of degree p with support starting at j.\n",
    "        Implemented recursively using the de Boor's recursion formula\"\"\"\n",
    "        cdef double w, w1\n",
    "        if p == 0:\n",
    "            if j == 0:\n",
    "                return 1.0\n",
    "            else:\n",
    "                return 0.0\n",
    "        else:\n",
    "            w = (x - j) / p\n",
    "            w1 = (x - j - 1) / p\n",
    "        return w * bspline_cython(p-1,j,x)+(1-w1)*bspline_cython(p-1,j+1,x)\n",
    "\n",
    "class BSplineCython:\n",
    "    \n",
    "    def __init__(self, p, xmin, xmax, ncells):\n",
    "        self.p = p\n",
    "        self.ncells = ncells\n",
    "        # compute eigenvalues of degree p b-spline matrix\n",
    "        self.modes = 2 * np.pi * np.arange(ncells) / ncells\n",
    "        self.deltax = (xmax - xmin) / ncells\n",
    "        \n",
    "        self.eig_bspl = bspline_cython(p,-(p+1)//2, 0.0)\n",
    "        for j in range(1, (p + 1) // 2):\n",
    "            self.eig_bspl += bspline_cython(p,j-(p+1)//2,0.0)*2*np.cos(j*self.modes)\n",
    "            \n",
    "        self.eigalpha = np.zeros(ncells, dtype=complex)\n",
    "    \n",
    "    @cython.boundscheck(False)\n",
    "    @cython.wraparound(False)\n",
    "    def __call__(self,  f,  alpha):\n",
    "        \"\"\"compute the interpolating spline of degree p of odd degree \n",
    "        of a function f on a periodic uniform mesh, at\n",
    "        all points xi-alpha\"\"\"\n",
    "        cdef Py_ssize_t j\n",
    "        cdef int p = self.p\n",
    "        # compute eigenvalues of cubic splines evaluated at displaced points\n",
    "        cdef int ishift = np.floor(-alpha / self.deltax)\n",
    "        cdef double beta = -ishift - alpha / self.deltax\n",
    "        self.eigalpha.fill(0)\n",
    "        for j in range(-(p-1)//2, (p+1)//2+1):\n",
    "            self.eigalpha += bspline_cython(p,j-(p+1)//2,beta)*np.exp((ishift+j)*1j*self.modes)\n",
    "            \n",
    "        # compute interpolating spline using fft and properties of circulant matrices\n",
    "        return np.real(ifft(fft(f) * self.eigalpha / self.eig_bspl))\n",
    "\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "times[\"cython\"] = []\n",
    "for M in tqdm(Mrange):\n",
    "    x = np.linspace(0, 1, M, endpoint=False)\n",
    "    f = np.sin(x*4*np.pi)\n",
    "    cs = BSplineCython(5, 0, 1, M )\n",
    "    \n",
    "    alpha = 0.1\n",
    "    ts = %timeit -oq -n 100 cs(f, alpha)\n",
    "    times[\"cython\"].append(ts.best)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "for o, t in times.items():\n",
    "    plt.loglog(Mrange, t, label=o)\n",
    "    \n",
    "plt.legend(loc='lower right')\n",
    "plt.xlabel('Number of points')\n",
    "plt.ylabel('Execution Time (s)');"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Vlasov-Poisson system\n",
    "system for one species\n",
    "with a neutralizing background.\n",
    "\n",
    "$$ \n",
    "\\frac{\\partial f}{\\partial t}+ v\\cdot \\nabla_x f + E(t,x) \\cdot \\nabla_v f = 0, \\\\\n",
    "- \\Delta \\phi = 1 - \\rho, E = - \\nabla \\phi \\\\\n",
    "\\rho(t,x)  =  \\int f(t,x,v)dv.\n",
    "$$\n",
    "\n",
    "- [Vlasov Equation - Wikipedia](https://en.wikipedia.org/wiki/Vlasov_equation)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class VlasovPoisson:\n",
    "    \n",
    "    def __init__(self, xmin, xmax, nx, vmin, vmax, nv, p = 3, opt=None):\n",
    "        \n",
    "        # Grid\n",
    "        self.nx = nx\n",
    "        self.x, self.dx = np.linspace(xmin, xmax, nx, endpoint=False, retstep=True)\n",
    "        self.nv = nv\n",
    "        self.v, self.dv = np.linspace(vmin, vmax, nv, endpoint=False, retstep=True)\n",
    "        \n",
    "        # Distribution function\n",
    "        self.f = np.zeros((nx,nv))\n",
    "        \n",
    "        self.p = p\n",
    "        \n",
    "        if opt :\n",
    "            # Interpolators for advection\n",
    "            self.cs_x = Advection(p, xmin, xmax, nx, opt)\n",
    "            self.cs_v = Advection(p, vmin, vmax, nv, opt)\n",
    "            \n",
    "        # Modes for Poisson equation\n",
    "        self.modes = np.zeros(nx)\n",
    "        k =  2* np.pi / (xmax - xmin)\n",
    "        self.modes[:nx//2] = k * np.arange(nx//2)\n",
    "        self.modes[nx//2:] = - k * np.arange(nx//2,0,-1)\n",
    "        self.modes += self.modes == 0 # avoid division by zero \n",
    "        \n",
    "    def advection_x(self, dt):\n",
    "        for j in range(self.nv):\n",
    "            alpha = dt * self.v[j]\n",
    "            self.f[j,:] = self.cs_x(self.f[j,:], alpha)\n",
    "            \n",
    "    def advection_v(self, e, dt):\n",
    "        for i in range(self.nx):\n",
    "            alpha = dt * e[i] \n",
    "            self.f[:,i] = self.cs_v(self.f[:,i], alpha)\n",
    "            \n",
    "    def compute_rho(self):\n",
    "        rho = self.dv * np.sum(self.f, axis=0)\n",
    "        return  rho - rho.mean()\n",
    "            \n",
    "    def compute_e(self, rho):\n",
    "        # compute Ex using that ik*Ex = rho\n",
    "        rhok = fft(rho)/self.modes\n",
    "        return np.real(ifft(-1j*rhok))\n",
    "    \n",
    "    def run(self, f, nstep, dt):\n",
    "        self.f = f\n",
    "        nrj = []\n",
    "        self.advection_x(0.5*dt)\n",
    "        for istep in tqdm(range(nstep)):\n",
    "            rho = self.compute_rho()\n",
    "            e = self.compute_e(rho)\n",
    "            self.advection_v(e, dt)\n",
    "            self.advection_x(dt)\n",
    "            nrj.append( 0.5*np.log(np.sum(e*e)*self.dx))\n",
    "                \n",
    "        return nrj"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Landau Damping\n",
    "\n",
    "[Landau damping - Wikipedia](https://en.wikipedia.org/wiki/Landau_damping)\n",
    "\n",
    "$$\n",
    "f(x, v, t=0) = \\frac{1}{\\sqrt{2\\pi}} (1 + \\epsilon  \\cdot cos(k_x \\cdot x))  e^{-v^2/2}\n",
    "$$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from time import time\n",
    "\n",
    "elapsed_time = {}\n",
    "fig, axes = plt.subplots()\n",
    "for opt in opts:\n",
    "    \n",
    "    # Set grid\n",
    "    nx, nv = 32, 64\n",
    "    xmin, xmax = 0.0, 4*np.pi\n",
    "    vmin, vmax = -6., 6.\n",
    "    \n",
    "    # Create Vlasov-Poisson simulation\n",
    "    sim = VlasovPoisson(xmin, xmax, nx, vmin, vmax, nv, opt=opt)\n",
    "\n",
    "    # Initialize distribution function\n",
    "    X, V = np.meshgrid(sim.x, sim.v)\n",
    "    eps, kx = 0.001, 0.5\n",
    "    f = (1.0+eps*np.cos(kx*X))/np.sqrt(2.0*np.pi)* np.exp(-0.5*V*V)\n",
    "\n",
    "    # Set time domain\n",
    "    nstep = 600\n",
    "    t, dt = np.linspace(0.0, 60.0, nstep, retstep=True)\n",
    "    \n",
    "    # Run simulation\n",
    "    etime = time()\n",
    "    nrj = sim.run(f, nstep, dt)\n",
    "    print(\" {0:12s} : {1:.4f} \".format(opt.__name__, time()-etime))\n",
    "    \n",
    "    # Plot energy\n",
    "    axes.plot(t, nrj, label=opt)\n",
    "\n",
    "    \n",
    "axes.plot(t, -0.1533*t-5.50)\n",
    "plt.legend();"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%%fortran --f90flags \"-fopenmp\" --extra \"-lgomp\" --link fftw3\n",
    "\n",
    "module bsl_fftw\n",
    "\n",
    "    use, intrinsic :: iso_c_binding\n",
    "    implicit none\n",
    "    include 'fftw3.f03'\n",
    "    \n",
    "    contains\n",
    "    \n",
    "   \n",
    "    recursive function bspline(p, j, x) result(res)\n",
    "    \n",
    "        integer :: p, j\n",
    "        real(8) :: x, w, w1\n",
    "        real(8) :: res\n",
    "     \n",
    "        if (p == 0) then\n",
    "            if (j == 0) then\n",
    "                res = 1.0\n",
    "                return\n",
    "            else\n",
    "                res = 0.0\n",
    "                return\n",
    "            end if\n",
    "        else\n",
    "            w = (x - j) / p\n",
    "            w1 = (x - j - 1) / p\n",
    "        end if\n",
    "     \n",
    "        res = w * bspline(p-1,j,x)+(1-w1)*bspline(p-1,j+1,x)\n",
    "    \n",
    "    end function bspline\n",
    "   \n",
    "    subroutine advection(p, n, delta, alpha, axis, df)\n",
    "    \n",
    "        integer, intent(in)     :: p\n",
    "        integer, intent(in)     :: n\n",
    "        real(8), intent(in)     :: delta\n",
    "        real(8), intent(in)     :: alpha(0:n-1)\n",
    "        integer, intent(in)     :: axis\n",
    "        real(8), intent(inout)  :: df(0:n-1,0:n-1)\n",
    "              \n",
    "        !f2py optional , depend(in) :: n=shape(df,0)\n",
    "     \n",
    "        real(8),    allocatable :: f(:)\n",
    "        complex(8), allocatable :: ft(:)\n",
    "        real(8),    allocatable :: eig_bspl(:)\n",
    "        real(8),    allocatable :: modes(:)\n",
    "        complex(8), allocatable :: eigalpha(:)\n",
    "        integer(8)              :: fwd\n",
    "        integer(8)              :: bwd\n",
    "     \n",
    "        integer                 :: i\n",
    "        integer                 :: j\n",
    "        integer                 :: ishift\n",
    "        real(8)                 :: beta\n",
    "        real(8)                 :: pi \n",
    "              \n",
    "        pi = 4.0_8 * atan(1.0_8)\n",
    "    \n",
    "        allocate(modes(0:n/2))\n",
    "        allocate(eigalpha(0:n/2))\n",
    "        allocate(eig_bspl(0:n/2))\n",
    "        allocate(f(0:n-1))\n",
    "        allocate(ft(0:n/2))\n",
    "     \n",
    "        do i = 0, n/2\n",
    "           modes(i) = 2.0_8 * pi * i / n\n",
    "        end do\n",
    "      \n",
    "        call dfftw_plan_dft_r2c_1d(fwd, n, f, ft, FFTW_ESTIMATE)\n",
    "        call dfftw_plan_dft_c2r_1d(bwd, n, ft, f, FFTW_ESTIMATE)\n",
    "     \n",
    "        eig_bspl = 0.0   \n",
    "        do j = 1, (p+1)/2\n",
    "           eig_bspl = eig_bspl + bspline(p, j-(p+1)/2, 0.0_8) * 2.0 * cos(j * modes)\n",
    "        end do\n",
    "     \n",
    "        eig_bspl = eig_bspl + bspline(p, -(p+1)/2, 0.0_8)\n",
    "      \n",
    "        !$OMP PARALLEL DO DEFAULT(FIRSTPRIVATE), SHARED(df,fwd,bwd)\n",
    "        do i = 0, n-1\n",
    "     \n",
    "            ishift = floor(-alpha(i) / delta)\n",
    "            beta   = -ishift - alpha(i) / delta\n",
    "            eigalpha = (0.0_8, 0.0_8)\n",
    "          \n",
    "            do j=-(p-1)/2, (p+1)/2+1\n",
    "               eigalpha = eigalpha + bspline(p, j-(p+1)/2, beta) * exp((ishift+j) * (0.0_8,1.0_8) * modes)\n",
    "            end do     \n",
    "            \n",
    "            if (axis == 0) then\n",
    "              f = df(:,i)\n",
    "            else\n",
    "              f = df(i,:)\n",
    "            end if\n",
    "   \n",
    "            call dfftw_execute_dft_r2c(fwd, f, ft)\n",
    "            ft = ft * eigalpha / eig_bspl / n\n",
    "            call dfftw_execute_dft_c2r(bwd, ft, f)\n",
    "           \n",
    "            if (axis == 0) then\n",
    "              df(:,i) = f\n",
    "            else\n",
    "              df(i,:) = f\n",
    "            end if\n",
    "  \n",
    "        end do\n",
    "        !$OMP END PARALLEL DO     \n",
    "        \n",
    "        call dfftw_destroy_plan(fwd)\n",
    "        call dfftw_destroy_plan(bwd)\n",
    "        deallocate(modes)\n",
    "        deallocate(eigalpha)\n",
    "        deallocate(eig_bspl)\n",
    "        deallocate(f)\n",
    "        deallocate(ft)\n",
    "    \n",
    "    end subroutine advection\n",
    "\n",
    "\n",
    "end module bsl_fftw"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%env OMP_NUM_THREADS=4"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from tqdm.notebook import tqdm\n",
    "from scipy.fftpack import fft, ifft\n",
    "\n",
    "class VlasovPoissonThreaded(VlasovPoisson):\n",
    "    \n",
    "    def advection_x(self, dt):\n",
    "        alpha = dt * self.v\n",
    "        bsl_fftw.advection(self.p, self.dx, alpha, 1, self.f)        \n",
    "            \n",
    "    def advection_v(self, e, dt):\n",
    "        alpha = dt * e\n",
    "        bsl_fftw.advection(self.p, self.dv, alpha, 0, self.f)\n",
    "            "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from time import time\n",
    "\n",
    "elapsed_time = {}\n",
    "fig, axes = plt.subplots()\n",
    "# Set grid\n",
    "nx, nv = 64, 64\n",
    "xmin, xmax = 0.0, 4*np.pi\n",
    "vmin, vmax = -6., 6.\n",
    "\n",
    "# Create Vlasov-Poisson simulation\n",
    "sim = VlasovPoissonThreaded(xmin, xmax, nx, vmin, vmax, nv)\n",
    "# Initialize distribution function\n",
    "X, V = np.meshgrid(sim.x, sim.v)\n",
    "eps, kx = 0.001, 0.5\n",
    "f = (1.0+eps*np.cos(kx*X))/np.sqrt(2.0*np.pi)* np.exp(-0.5*V*V)\n",
    "f = np.asfortranarray(f)\n",
    "# Set time domain\n",
    "nstep = 600\n",
    "t, dt = np.linspace(0.0, 60.0, nstep, retstep=True)\n",
    "\n",
    "# Run simulation\n",
    "etime = time()\n",
    "nrj = sim.run(f, nstep, dt)\n",
    "print(\" elapsed time : {0:.4f} \".format(time()-etime))\n",
    "\n",
    "# Plot energy\n",
    "axes.plot(t, nrj, label='energy')\n",
    "\n",
    "    \n",
    "plt.legend();"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
